/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (C) 2010-2012 Ken Tossell
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the author nor other contributors may be
 *     used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

/**
 * @defgroup test Example code of test application
 * @brief Capture video frame of UVC device, and render in OpenCV window
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include "libuvc/libuvc_hx.h"
#include "SaveXyz.h"

#define ENABLE_NIR_SENSOR   1
#define ENABLE_RGB_SENSOR   1
#define VGA                 0 /* Enable if using specific romcode*/

#if ENABLE_NIR_SENSOR
#define USB_VID             (0x1080)
#define USB_PID             (0x1063)
#if VGA
#define NIR_CAPTURE_WIDTH   (640)
#define NIR_CAPTURE_HEIGHT  (401)
#define NIR_CAPTURE_FPS     (30)
#define SCALING             (1)
#else
#define NIR_CAPTURE_WIDTH   (1280)
#define NIR_CAPTURE_HEIGHT  (801)
#define NIR_CAPTURE_FPS     (30)
#define SCALING             (0)
#endif // VGA
#define NIR_CAPTURE_FORMAT  UVC_FRAME_FORMAT_YUYV
#endif // ENABLE_NIR_SENSOR

#if ENABLE_RGB_SENSOR
#define RGB_USB_VID         (0x1080)
#define RGB_USB_PID         (0x1245)
#if VGA
#define RGB_CAPTURE_WIDTH   (640)
#define RGB_CAPTURE_HEIGHT  (481)
#define RGB_CAPTURE_FPS     (30)
#else
#define RGB_CAPTURE_WIDTH   (1280)
#define RGB_CAPTURE_HEIGHT  (801)
#define RGB_CAPTURE_FPS     (30)
#endif // VGA
#define RGB_CAPTURE_FORMAT  UVC_FRAME_FORMAT_YUYV
#endif // ENABLE_RGB_SENSOR

#define ENABLE_DUMP_IMAGE     (0)
#define DUMP_IMAGE_MAX_COUNT  (10)

#define CX3_FW_VERSION_LEN    (16)

/*  Macros to cut down on compiler warnings. */
#define UNUSED(expr) \
  do                 \
  {                  \
    (void)(expr);    \
  } while (0)

static int non_stop = 1;

/* variables and macros for socket connection*/
static char const *ip = "127.0.0.1";
typedef enum port_number
{
  nir_port = 8000,
  depth_port = 8001,
  dot_port = 8002,
  rgb_port = 8003,
} port_number_t;

/**
 * @brief Dump frame buffer to file.
 * @ingroup debug
 */
#if ENABLE_DUMP_IMAGE
void save_img(const void *buf, size_t len, const char *fname_prefix, int fname_idx)
{
  char fname[100] = {0};
  FILE *fp = NULL;

  snprintf(fname, 100, "%s_%02d.raw", fname_prefix, fname_idx);
  if ((fp = fopen(fname, "w+")) != NULL)
  {
    fwrite(buf, len, 1, fp);
    fclose(fp);
    printf("Saved image: %s\n", fname);
  }
}
#else
void save_img(const void *buf, size_t len, const char *fname_prefix, int fname_idx)
{
  UNUSED(buf);
  UNUSED(len);
  UNUSED(fname_prefix);
  UNUSED(fname_idx);
}
#endif // ENABLE_DUMP_IMAGE

/**
 * @brief SIGINT signal handler.
 * @ingroup test
 */
void sigint(int a)
{
  UNUSED(a);
  printf("\naborting...\n");
  non_stop = 0;
}

/**
 * @brief Set socket connection to host server
 * @ingroup
 */
static int init_tcp(int port)
{
  struct sockaddr_in server_addr;

  int ret;
  int sockfd;
  sockfd = socket(AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0)
    return -1;

  const int enable = 1;
  setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable));

  server_addr.sin_family = AF_INET;
  server_addr.sin_port = htons(port);
  server_addr.sin_addr.s_addr = inet_addr(ip);

  ret = connect(sockfd, (struct sockaddr *)&server_addr, sizeof(server_addr));
  if (ret < 0)
  {
    close(sockfd);
    return -1;
  }

  return sockfd;
}

#if ENABLE_NIR_SENSOR
/**
 * @brief NIR frame callback function.
 * @ingroup test
 */
void nir_cb(uvc_frame_t *frame, void *ptr)
{
  uvc_frame_t *y16_r270 = NULL;
  uvc_frame_t *y10_r270 = NULL;
  uvc_frame_t *y8_r270 = NULL;
  uvc_frame_t *bgr_r270 = NULL;
  uvc_error_t ret;
  static int cnt_nir_raw = 1;
  static int cnt_nir_y16_r270 = 1;
  static int cnt_nir_y10_r270 = 1;
  static int cnt_nir_y8_r270 = 1;
  static int cnt_nir_bgr888_r270 = 1;
  static int sockfd;
  static int rest_package_size;
  static int send_size = 0;
  static int port = nir_port;

  UNUSED(ptr);

  if (!frame)
  {
    printf("NIR frame nullptr!\n");
    goto err;
  }

  if (cnt_nir_raw <= DUMP_IMAGE_MAX_COUNT)
    save_img(frame->data, frame->data_bytes, "img_nir_raw", cnt_nir_raw++);

  y16_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y16_r270)
  {
    printf("NIR: unable to allocate y16_r270 frame!\n");
    goto err;
  }

  y10_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y10_r270)
  {
    printf("NIR: unable to allocate y10_r270 frame!\n");
    goto err;
  }

  y8_r270 = uvc_allocate_frame(frame->width * frame->height * 1);
  if (!y8_r270)
  {
    printf("NIR: unable to allocate y8_r270 frame!\n");
    goto err;
  }

  bgr_r270 = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr_r270)
  {
    printf("NIR: unable to allocate bgr_r270 frame!\n");
    goto err;
  }

  ret = uvc_hx_y16_rotate(frame, y16_r270, UVC_HX_ROTATION_270);
  if (ret != UVC_SUCCESS)
  {
    uvc_perror(ret, "NIR: uvc_hx_y16_rotate(270)");
    goto err;
  }

  if (cnt_nir_y16_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y16_r270->data, y16_r270->data_bytes, "img_nir_y16_r270", cnt_nir_y16_r270++);

  // Left-aligned => right-aligned 10-bit
  ret = uvc_y16_shr(y16_r270, y10_r270, 6);
  if (ret != UVC_SUCCESS)
  {
    uvc_perror(ret, "NIR: uvc_y16_shr(6)");
    goto err;
  }

  if (cnt_nir_y10_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y10_r270->data, y10_r270->data_bytes, "img_nir_y10_r270", cnt_nir_y10_r270++);

  ret = uvc_y16_to_y8(y16_r270, y8_r270, HX_EFFECTIVE_BIT_ALIGNMENT_LEFT);
  if (ret)
  {
    uvc_perror(ret, "NIR: uvc_y16_to_y8");
    goto err;
  }

  if (cnt_nir_y8_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y8_r270->data, y8_r270->data_bytes, "img_nir_y8_r270", cnt_nir_y8_r270++);

  ret = uvc_any2bgr(y8_r270, bgr_r270);
  if (ret)
  {
    uvc_perror(ret, "NIR: uvc_any2bgr");
    goto err;
  }

  if (cnt_nir_bgr888_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr_r270->data, bgr_r270->data_bytes, "img_nir_bgr888_r270", cnt_nir_bgr888_r270++);

  /* Send frame via tcp*/
  sockfd = init_tcp(port);
  if (sockfd < 0)
    goto err;

  rest_package_size = bgr_r270->data_bytes;
  while (rest_package_size > 0)
  {
    send_size = send(sockfd, bgr_r270->data, bgr_r270->data_bytes, 0);
    if (send_size <= 0)
      goto err;
    rest_package_size = rest_package_size - send_size;
  }
  close(sockfd);

err:
  if (bgr_r270)
    uvc_free_frame(bgr_r270);
  if (y8_r270)
    uvc_free_frame(y8_r270);
  if (y10_r270)
    uvc_free_frame(y10_r270);
  if (y16_r270)
    uvc_free_frame(y16_r270);
}

/**
 * @brief Depth frame callback function.
 * @ingroup test
 */
void depth_cb(uvc_frame_t *frame, void *ptr)
{
  uvc_frame_t *y16_r270 = NULL;
  uvc_frame_t *q12_r270 = NULL;
  uvc_frame_t *bgr_r270 = NULL;
  uvc_error_t ret;
  static int cnt_dep_raw = 1;
  static int cnt_dep_y16_r270 = 1;
  static int cnt_dep_q12_r270 = 1;
  static int cnt_dep_bgr888_r270 = 1;
  static int sockfd;
  static int rest_package_size;
  static int send_size = 0;
  static int port = depth_port;

  UNUSED(ptr);

  if (!frame)
  {
    printf("Depth frame nullptr!\n");
    goto err;
  }

  // Test: Save XYZ
  if (cnt_dep_raw <= DUMP_IMAGE_MAX_COUNT)
    SaveXyz(frame->data, cnt_dep_raw);

  if (cnt_dep_raw <= DUMP_IMAGE_MAX_COUNT)
    save_img(frame->data, frame->data_bytes, "img_dep_raw", cnt_dep_raw++);

  y16_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y16_r270)
  {
    printf("Depth: unable to allocate y16_r270 frame!\n");
    goto err;
  }

  q12_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!q12_r270)
  {
    printf("Depth: unable to allocate q12_r270 frame!\n");
    goto err;
  }

  bgr_r270 = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr_r270)
  {
    printf("Depth: unable to allocate bgr_r270 frame!\n");
    goto err;
  }

  ret = uvc_hx_y16_rotate(frame, y16_r270, UVC_HX_ROTATION_270);
  if (ret != UVC_SUCCESS)
  {
    uvc_perror(ret, "Depth: uvc_hx_y16_rotate(270)\n");
    goto err;
  }

  if (cnt_dep_y16_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y16_r270->data, y16_r270->data_bytes, "img_dep_y16_r270", cnt_dep_y16_r270++);

  // Q12.4 => Q12
  ret = uvc_y16_shr(y16_r270, q12_r270, 4);
  if (ret != UVC_SUCCESS)
  {
    uvc_perror(ret, "Depth: uvc_y16_shr(4)");
    goto err;
  }

  if (cnt_dep_q12_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(q12_r270->data, q12_r270->data_bytes, "img_dep_q12_r270", cnt_dep_q12_r270++);

  ret = uvc_y16_to_depth_colormap_bgr(q12_r270, bgr_r270, 1);
  if (ret != UVC_SUCCESS)
  {
    uvc_perror(ret, "Depth: uvc_y16_to_depth_colormap_bgr");
    goto err;
  }

  if (cnt_dep_bgr888_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr_r270->data, bgr_r270->data_bytes, "img_dep_bgr888_r270", cnt_dep_bgr888_r270++);

  /* Send frame via tcp*/
  sockfd = init_tcp(port);
  if (sockfd < 0)
    goto err;

  rest_package_size = bgr_r270->data_bytes;
  while (rest_package_size > 0)
  {
    send_size = send(sockfd, bgr_r270->data, bgr_r270->data_bytes, 0);
    if (send_size <= 0)
      goto err;
    rest_package_size = rest_package_size - send_size;
  }
  close(sockfd);

err:
  if (bgr_r270)
    uvc_free_frame(bgr_r270);
  if (q12_r270)
    uvc_free_frame(q12_r270);
  if (y16_r270)
    uvc_free_frame(y16_r270);
}

/**
 * @brief Dot frame callback function.
 * @ingroup test
 */
void dot_cb(uvc_frame_t *frame, void *ptr)
{
  uvc_frame_t *y16_r270 = NULL;
  uvc_frame_t *y10_r270 = NULL;
  uvc_frame_t *y8_r270 = NULL;
  uvc_frame_t *bgr_r270 = NULL;
  uvc_error_t ret;
  static int cnt_dot_raw = 1;
  static int cnt_dot_y16_r270 = 1;
  static int cnt_dot_y10_r270 = 1;
  static int cnt_dot_y8_r270 = 1;
  static int cnt_dot_bgr888_r270 = 1;
  static int sockfd;
  static int rest_package_size;
  static int send_size = 0;
  static int port = dot_port;

  UNUSED(ptr);

  if (!frame)
  {
    printf("Dot frame nullptr!\n");
    goto err;
  }

  if (cnt_dot_raw <= DUMP_IMAGE_MAX_COUNT)
    save_img(frame->data, frame->data_bytes, "img_dot_raw", cnt_dot_raw++);

  y16_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y16_r270)
  {
    printf("Dot: unable to allocate y16_r270 frame!\n");
    goto err;
  }

  y10_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y10_r270)
  {
    printf("Dot: unable to allocate y10_r270 frame!\n");
    goto err;
  }

  y8_r270 = uvc_allocate_frame(frame->width * frame->height * 1);
  if (!y8_r270)
  {
    printf("Dot: unable to allocate y8_r270 frame!\n");
    goto err;
  }

  bgr_r270 = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr_r270)
  {
    printf("Dot: unable to allocate bgr_r270 frame!\n");
    goto err;
  }

  ret = uvc_hx_y16_rotate(frame, y16_r270, UVC_HX_ROTATION_270);
  if (ret != UVC_SUCCESS)
  {
    uvc_perror(ret, "Dot: uvc_hx_y16_rotate(270)\n");
    goto err;
  }

  if (cnt_dot_y16_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y16_r270->data, y16_r270->data_bytes, "img_dot_y16_r270", cnt_dot_y16_r270++);

  // Left-aligned => right-aligned 10-bit
  ret = uvc_y16_shr(y16_r270, y10_r270, 6);
  if (ret != UVC_SUCCESS)
  {
    uvc_perror(ret, "Dot: uvc_y16_shr(6)");
    goto err;
  }

  if (cnt_dot_y10_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y10_r270->data, y10_r270->data_bytes, "img_dot_y10_r270", cnt_dot_y10_r270++);

  ret = uvc_y16_to_y8(y16_r270, y8_r270, HX_EFFECTIVE_BIT_ALIGNMENT_LEFT);
  if (ret)
  {
    uvc_perror(ret, "Dot: uvc_y16_to_y8");
    goto err;
  }

  if (cnt_dot_y8_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y8_r270->data, y8_r270->data_bytes, "img_dot_y8_r270", cnt_dot_y8_r270++);

  ret = uvc_any2bgr(y8_r270, bgr_r270);
  if (ret)
  {
    uvc_perror(ret, "Dot: uvc_any2bgr");
    goto err;
  }

  if (cnt_dot_bgr888_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr_r270->data, bgr_r270->data_bytes, "img_dot_bgr888_r270", cnt_dot_bgr888_r270++);

  /* Send frame via tcp*/
  sockfd = init_tcp(port);
  if (sockfd < 0)
    goto err;

  rest_package_size = bgr_r270->data_bytes;
  while (rest_package_size > 0)
  {
    send_size = send(sockfd, bgr_r270->data, bgr_r270->data_bytes, 0);
    if (send_size <= 0)
      goto err;
    rest_package_size = rest_package_size - send_size;
  }
  close(sockfd);

err:
  if (bgr_r270)
    uvc_free_frame(bgr_r270);
  if (y8_r270)
    uvc_free_frame(y8_r270);
  if (y10_r270)
    uvc_free_frame(y10_r270);
  if (y16_r270)
    uvc_free_frame(y16_r270);
}
#endif // ENABLE_NIR_SENSOR

#if ENABLE_RGB_SENSOR
/**
 * @brief RGB Frame callback function.
 * @ingroup test
 */
void rgb_cb(uvc_frame_t *frame, void *ptr)
{
  uvc_frame_t *bgr = NULL;
  uvc_frame_t *bgr_r270 = NULL;
  uvc_error_t ret;
  static int cnt_rgb_yuv422 = 1;
  static int cnt_rgb_bgr888 = 1;
  static int cnt_rgb_bgr888_r270 = 1;
  static int sockfd;
  static int rest_package_size;
  static int send_size = 0;
  static int port = rgb_port;

  UNUSED(ptr);

  if (!frame)
  {
    printf("RGB frame nullptr!\n");
    goto err;
  }

  if (cnt_rgb_yuv422 <= DUMP_IMAGE_MAX_COUNT)
    save_img(frame->data, frame->data_bytes, "img_rgb_yuv422", cnt_rgb_yuv422++);

  bgr = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr)
  {
    printf("RGB: unable to allocate bgr frame!\n");
    goto err;
  }

  bgr_r270 = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr_r270)
  {
    printf("Dot: unable to allocate bgr_r270 frame!\n");
    goto err;
  }

  ret = uvc_any2bgr(frame, bgr);
  if (ret)
  {
    uvc_perror(ret, "RGB: uvc_any2bgr");
    goto err;
  }

  if (cnt_rgb_bgr888 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr->data, bgr->data_bytes, "img_rgb_bgr888", cnt_rgb_bgr888++);

  ret = uvc_hx_bgr_rotate(bgr, bgr_r270, UVC_HX_ROTATION_270);
  if (ret)
  {
    uvc_perror(ret, "RGB: uvc_hx_rgb_rotate(270)\n");
    goto err;
  }

  if (cnt_rgb_bgr888_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr_r270->data, bgr_r270->data_bytes, "img_rgb_bgr888_r270", cnt_rgb_bgr888_r270++);

  /* Send frame via tcp*/
  sockfd = init_tcp(port);
  if (sockfd < 0)
    goto err;

  rest_package_size = bgr_r270->data_bytes;
  while (rest_package_size > 0)
  {
    send_size = send(sockfd, bgr_r270->data, bgr_r270->data_bytes, 0);
    if (send_size <= 0)
      goto err;
    rest_package_size = rest_package_size - send_size;
  }
  close(sockfd);

err:
  if (bgr)
    uvc_free_frame(bgr);
  if (bgr_r270)
    uvc_free_frame(bgr_r270);
}
#endif // ENABLE_RGB_SENSOR

/**
 * @brief main function.
 * @ingroup test
 */
int main(int argc, char **argv)
{
#if ENABLE_NIR_SENSOR
  uvc_context_t *nir_ctx;
  uvc_error_t nir_ret;
  uvc_device_t *nir_dev;
  uvc_device_handle_t *nir_devh;
  uvc_stream_ctrl_t nir_ctrl;

  hx_hv2_fw_version_t hv2_fw_version;
  hx_cam_intrinsic_extrinsic_matrices_t cam_matrices;

  uint32_t serial_number;
  uint16_t build_type;
  uint8_t hv2_romcode_version[4];
  uint8_t current_stream_mode;

  char cx3_fw_version[CX3_FW_VERSION_LEN] = {0};
#endif // ENABLE_NIR_SENSOR

#if ENABLE_RGB_SENSOR
  uvc_context_t *rgb_ctx;
  uvc_error_t rgb_ret;
  uvc_device_t *rgb_dev;
  uvc_device_handle_t *rgb_devh;
  uvc_stream_ctrl_t rgb_ctrl;
#endif // ENABLE_RGB_SENSOR

  UNUSED(argc);
  UNUSED(argv);

/* Open UVC device and initialize */
#if ENABLE_NIR_SENSOR
  /* Init NIR Sensor */
  nir_ret = uvc_init(&nir_ctx, NULL);

  if (nir_ret < 0)
  {
    uvc_perror(nir_ret, "uvc_init");
    return nir_ret;
  }
  else
    printf("UVC initialized\n");

  nir_ret = uvc_find_device(nir_ctx, &nir_dev, USB_VID, USB_PID, NULL);

  if (nir_ret < 0)
  {
    uvc_perror(nir_ret, "uvc_find_device");
    return nir_ret;
  }
  else
    printf("NIR Device found\n");

  nir_ret = uvc_open(nir_dev, &nir_devh);
  if (nir_ret < 0)
  {
    uvc_perror(nir_ret, "uvc_open");
    return nir_ret;
  }
  else
    printf("NIR Device opened\n");

  //uvc_print_diag(devh, stderr);

  nir_ret = uvc_get_stream_ctrl_format_size(
      nir_devh, &nir_ctrl, NIR_CAPTURE_FORMAT,
      NIR_CAPTURE_WIDTH, NIR_CAPTURE_HEIGHT,
      NIR_CAPTURE_FPS);

  if (nir_ret < 0)
  {
    uvc_perror(nir_ret, "get_mode");
    return nir_ret;
  }
  else
    printf("Select NIR format: %dx%d (%d fps)\n",
           NIR_CAPTURE_WIDTH, NIR_CAPTURE_HEIGHT, NIR_CAPTURE_FPS);

  //uvc_print_stream_ctrl(&nir_ctrl, stderr);

#endif // ENABLE_NIR_SENSOR

#if ENABLE_RGB_SENSOR
  /* Init RGB Sensor */
  rgb_ret = uvc_init(&rgb_ctx, NULL);
  if (rgb_ret < 0)
  {
    uvc_perror(rgb_ret, "RGB uvc_init");
    return rgb_ret;
  }
  else
    printf("UVC initialized\n");

  rgb_ret = uvc_find_device(rgb_ctx, &rgb_dev, RGB_USB_VID, RGB_USB_PID, NULL);
  if (rgb_ret < 0)
  {
    uvc_perror(rgb_ret, "RGB uvc_find_device");
    return rgb_ret;
  }
  else
    printf("RGB Device found\n");

  rgb_ret = uvc_open(rgb_dev, &rgb_devh);
  if (rgb_ret < 0)
  {
    uvc_perror(rgb_ret, "RGB uvc_open");
    return rgb_ret;
  }
  else
    printf("RGB Device opened\n");

  //uvc_print_diag(rgb_devh, stderr);

  rgb_ret = uvc_get_stream_ctrl_format_size(
      rgb_devh, &rgb_ctrl, RGB_CAPTURE_FORMAT,
      RGB_CAPTURE_WIDTH, RGB_CAPTURE_HEIGHT,
      RGB_CAPTURE_FPS);

  if (rgb_ret < 0)
  {
    uvc_perror(rgb_ret, "RGB get_mode");
    return rgb_ret;
  }
  else
    printf("Select RGB format: %dx%d (%d fps)\n",
           RGB_CAPTURE_WIDTH, RGB_CAPTURE_HEIGHT, RGB_CAPTURE_FPS);

  //uvc_print_stream_ctrl(&rgb_ctrl, stderr);

#endif // ENABLE_RGB_SENSOR

#if ENABLE_NIR_SENSOR
  /* Read SH430MH Information */
  nir_ret = uvc_hx_get_cx3_fw_version(nir_devh, cx3_fw_version, sizeof(cx3_fw_version));
  if (nir_ret != UVC_SUCCESS)
    uvc_perror(nir_ret, "uvc_hx_get_cx3_fw_version failed");
  else
    printf("cx3_fw_version = %s\n", cx3_fw_version);

  nir_ret = uvc_hx_get_hv2_fw_version(nir_devh, &hv2_fw_version);
  if (nir_ret != UVC_SUCCESS)
    uvc_perror(nir_ret, "uvc_hx_get_hv2_fw_version failed");
  else
    printf("hv2_fw_version: chip_id=0x%04X, version_major='%c', "
           "version_date=%02u.%02u.%02u, version_minor=0x%02X, customer_id=0x%02X\n",
           hv2_fw_version.chip_id,
           hv2_fw_version.version_major,
           hv2_fw_version.version_date_year,
           hv2_fw_version.version_date_month,
           hv2_fw_version.version_date_day,
           hv2_fw_version.version_minor,
           hv2_fw_version.customer_id);

  nir_ret = uvc_hx_get_hv2_romcode_version(nir_devh, hv2_romcode_version, sizeof(hv2_romcode_version));
  if (nir_ret != UVC_SUCCESS)
    uvc_perror(nir_ret, "uvc_hx_get_hv2_romcode_version failed");
  else
    printf("hv2_romcode_version: %02X %02X %02X %02X\n",
           hv2_romcode_version[0],
           hv2_romcode_version[1],
           hv2_romcode_version[2],
           hv2_romcode_version[3]);

  build_type = 0;
  nir_ret = uvc_hx_get_build_type(nir_devh, &build_type);
  if (nir_ret != UVC_SUCCESS)
    uvc_perror(nir_ret, "Failed to get Himax camera build type");
  else
    printf("Himax camera build type = 0x%04X\n", build_type);

  serial_number = 0;
  nir_ret = uvc_hx_get_serial_number(nir_devh, &serial_number);
  if (nir_ret != UVC_SUCCESS)
    uvc_perror(nir_ret, "Failed to get Himax camera serial number");
  else
    printf("Himax camera serial number = 0x%08X (%u)\n", serial_number, serial_number);

  memset(&cam_matrices, 0, sizeof(hx_cam_intrinsic_extrinsic_matrices_t));
  nir_ret = uvc_hx_get_intrinsic_extrinsic_matrices(nir_devh, &cam_matrices, SCALING);
  if (nir_ret != UVC_SUCCESS)
    uvc_perror(nir_ret, "Failed to get Himax camera sensors' intrinsic extrinsic parameter matrices");
  else
  {
    printf("Himax camera sensors' intrinsic extrinsic parameter matrices:\n");
    nir_ret = uvc_hx_fprintf_intrinsic_extrinsic_matrices(stdout, &cam_matrices);
    if (nir_ret != UVC_SUCCESS)
      printf("Failed to print Himax camera sensors' intrinsic extrinsic parameter matrices\n");
  }
#endif // ENABLE_NIR_SENSOR

/* Start Streaming */
#if ENABLE_NIR_SENSOR
  nir_ret = uvc_start_streaming_multi_cb(nir_devh, &nir_ctrl,
                                         nir_cb, (void *)1001,
                                         depth_cb, (void *)1002,
                                         dot_cb, (void *)1003,
                                         0);
  if (nir_ret < 0)
  {
    uvc_perror(nir_ret, "NIR start_streaming");
    return nir_ret;
  }
#endif // ENABLE_NIR_SENSOR

#if ENABLE_RGB_SENSOR
  rgb_ret = uvc_start_streaming(rgb_devh, &rgb_ctrl, rgb_cb, (void *)67890, 0);
  if (rgb_ret < 0)
  {
    uvc_perror(rgb_ret, "RGB start_streaming");
    return rgb_ret;
  }
#endif // ENABLE_RGB_SENSOR

  /* Set Frame Mode */
#if ENABLE_NIR_SENSOR
  // nir_ret = uvc_hx_set_frame_mode(nir_devh, HX_FRAME_MODE_SELECT_ONLY_NIR);
  // nir_ret = uvc_hx_set_frame_mode(nir_devh, HX_FRAME_MODE_SELECT_ONLY_DOT);
  // nir_ret = uvc_hx_set_frame_mode(nir_devh, HX_FRAME_MODE_SELECT_ONLY_DEPTH);
  nir_ret = uvc_hx_set_frame_mode(nir_devh, HX_FRAME_MODE_SELECT_ALT_NIR_DEPTH);
  // nir_ret = uvc_hx_set_frame_mode(nir_devh, HX_FRAME_MODE_SELECT_ALT_NIR_DOT);
  // nir_ret = uvc_hx_set_frame_mode(nir_devh, HX_FRAME_MODE_SELECT_ALT_NIR_DEPTH_DOT);
  if (nir_ret != UVC_SUCCESS)
    uvc_perror(nir_ret, "uvc_hx_set_frame_mode failed");

  usleep(50000);

  nir_ret = uvc_hx_get_frame_mode(nir_devh, &current_stream_mode);
  if (nir_ret != UVC_SUCCESS)
    uvc_perror(nir_ret, "uvc_hx_get_frame_mode failed");
  else
    printf("current stream mode = %u\n", current_stream_mode);
#endif // ENABLE_NIR_SENSOR

#if ENABLE_NIR_SENSOR || ENABLE_RGB_SENSOR
  printf("Streaming... wait for pressing [Ctrl-C] to terminate\n");
  signal(SIGINT, sigint);

  while (non_stop)
  {
    /* Keep Streaming */
    printf("streaming...\n");
    sleep(5);
  }
#else
  printf("Please set [ENABLE_NIR_SENSOR] or [ENABLE_RGB_SENSOR]\n");
#endif

#if ENABLE_NIR_SENSOR
  uvc_stop_streaming(nir_devh);
  printf("Done NIR streaming\n");
  uvc_close(nir_devh);
  printf("NIR Device closed\n");
  uvc_unref_device(nir_dev);
  uvc_exit(nir_ctx);
  printf("NIR UVC exited\n");
#endif // ENABLE_NIR_SENSOR

#if ENABLE_RGB_SENSOR
  uvc_stop_streaming(rgb_devh);
  printf("Done RGB streaming\n");
  uvc_close(rgb_devh);
  printf("RGB Device closed\n");
  uvc_unref_device(rgb_dev);
  uvc_exit(rgb_ctx);
  printf("RGB UVC exited\n");
#endif // ENABLE_RGB_SENSOR

  return 0;
}
