#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <opencv2/highgui/highgui_c.h>

#define VGA                 0

/* The frame is rotated as portrait in client side.
 *
 * Frame size
 * */
#if VGA
#define NIR_SENSOR_WIDTH    (400)
#define NIR_SENSOR_HEIGHT   (640)
#define RGB_SENSOR_WIDTH    (480)
#define RGB_SENSOR_HEIGHT   (640)
#else
#define NIR_SENSOR_WIDTH    (800)
#define NIR_SENSOR_HEIGHT   (1280)
#define RGB_SENSOR_WIDTH    (800)
#define RGB_SENSOR_HEIGHT   (1280)
#endif
#define NIR_SENSOR_FRAME_SIZE     (NIR_SENSOR_WIDTH * NIR_SENSOR_HEIGHT * 3)
#define RGB_SENSOR_FRAME_SIZE     (RGB_SENSOR_WIDTH * RGB_SENSOR_HEIGHT * 3)
#define IMAGE_MAXIMUM_BUFFER_SIZE (RGB_SENSOR_FRAME_SIZE)

/* Socket */
static char const *ip = "127.0.0.1";
typedef enum port_number
{
  nir_port = 8000,
  depth_port = 8001,
  dot_port = 8002,
  rgb_port = 8003,
} port_number_t;

typedef struct socket_arg
{
  int new_socket_fd;
  struct sockaddr_in client_address;
} socket_arg_t;

static bool display_nir = false;
static bool display_depth = false;
static bool display_dot = false;
static bool display_rgb = false;

/* Tmp data*/
static uint8_t *nir_tmp;
static uint8_t *depth_tmp;
static uint8_t *dot_tmp;
static uint8_t *rgb_tmp;

/* Flag */
static pthread_mutex_t nir_lock;
static pthread_mutex_t depth_lock;
static pthread_mutex_t dot_lock;
static pthread_mutex_t rgb_lock;

static int non_stop = 1;

void recv_data(int sockfd, uint8_t *tmp_buffer, int port)
{
  uint8_t *buffer;
  int tmp_len = 0;
  int recv_buf_size;
  int recv_len, write_leng;
  int rest_package;

  if (port == htons(rgb_port))
  {
    buffer = (uint8_t *)malloc(RGB_SENSOR_FRAME_SIZE * sizeof(uint8_t));
    rest_package = RGB_SENSOR_FRAME_SIZE;
    recv_buf_size = RGB_SENSOR_FRAME_SIZE;
  }
  else
  {
    buffer = (uint8_t *)malloc(NIR_SENSOR_FRAME_SIZE * sizeof(uint8_t));
    rest_package = NIR_SENSOR_FRAME_SIZE;
    recv_buf_size = NIR_SENSOR_FRAME_SIZE;
  }

  while (rest_package > 0)
  {
    recv_len = recv(sockfd, buffer, recv_buf_size, 0);
    if (recv_len <= 0)
      break;

    memcpy(&tmp_buffer[tmp_len], buffer, recv_len);
    rest_package = rest_package - recv_len;
    tmp_len += recv_len;
  }
  free(buffer);
}

void signal_handler(int sig)
{
  non_stop = 0;
  printf("Close server socket successfully\n");
}

void *init_server_socket(void *server_info)
{
  struct sockaddr_in server_addr, new_addr;

  struct sockaddr_in *sock_info = (struct sockaddr_in *)server_info;

  int ret = 0;

  int server_socket = 0;
  server_socket = socket(AF_INET, SOCK_STREAM, 0);

  if (server_socket < 0)
  {
    perror("Error in socket");
    exit(EXIT_FAILURE);
  }

  const int enable = 1;
  setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable));

  printf("Server socket created successfully .\n");

  server_addr.sin_family = sock_info->sin_family;
  server_addr.sin_port = sock_info->sin_port;
  server_addr.sin_addr.s_addr = sock_info->sin_addr.s_addr;
  ret = bind(server_socket, (struct sockaddr *)&server_addr, sizeof(server_addr));
  if (ret < 0)
  {
    perror("Error in bind");
    exit(EXIT_FAILURE);
  }
  printf("Binding successfull.\n");

  if (listen(server_socket, 10) == 0)
  {
    printf("Listening....\n");
  }
  else
  {
    perror("Error in listening");
    exit(EXIT_FAILURE);
  }

  signal(SIGINT, signal_handler);

  printf("Waiting for incoming connections...\n");

  socket_arg_t *socket_arg;
  socklen_t client_addr_size;
  uint8_t *img_buff;
  int new_sock;
  while (non_stop)
  {
    socket_arg = (socket_arg_t *)malloc(sizeof *socket_arg);
    if (!socket_arg)
    {
      perror("malloc");
      continue;
    }

    client_addr_size = sizeof(new_addr);

    socket_arg->new_socket_fd = accept(server_socket, (struct sockaddr *)&socket_arg->client_address, &client_addr_size);

    if (socket_arg->new_socket_fd < 0)
    {
      perror("Accept socket failed\n");
      continue;
    }

    img_buff = (uint8_t *)malloc(IMAGE_MAXIMUM_BUFFER_SIZE * sizeof(uint8_t));
    recv_data(socket_arg->new_socket_fd, img_buff, server_addr.sin_port);

    if (server_addr.sin_port == htons(nir_port) && display_nir == false)
    {
      pthread_mutex_lock(&nir_lock);
      memcpy(nir_tmp, img_buff, NIR_SENSOR_FRAME_SIZE);
      display_nir = true;
      pthread_mutex_unlock(&nir_lock);
    }
    else if (server_addr.sin_port == htons(depth_port) && display_depth == false)
    {
      pthread_mutex_lock(&depth_lock);
      memcpy(depth_tmp, img_buff, NIR_SENSOR_FRAME_SIZE);
      display_depth = true;
      pthread_mutex_unlock(&depth_lock);
    }
    else if (server_addr.sin_port == htons(dot_port) && display_dot == false)
    {
      pthread_mutex_lock(&dot_lock);
      memcpy(dot_tmp, img_buff, NIR_SENSOR_FRAME_SIZE);
      display_dot = true;
      pthread_mutex_unlock(&dot_lock);
    }
    else if (server_addr.sin_port == htons(rgb_port) && display_rgb == false)
    {
      pthread_mutex_lock(&rgb_lock);
      memcpy(rgb_tmp, img_buff, RGB_SENSOR_FRAME_SIZE);
      display_rgb = true;
      pthread_mutex_unlock(&rgb_lock);
    }
    close(socket_arg->new_socket_fd);
    free(img_buff);
    free(socket_arg);
  }

  close(server_socket);
  exit(EXIT_SUCCESS);
}

void display_nir_img(uint8_t *img_frame, const char *window_name)
{
  IplImage *cvImg = cvCreateImageHeader(cvSize(NIR_SENSOR_WIDTH, NIR_SENSOR_HEIGHT), IPL_DEPTH_8U, 3);
  cvSetData(cvImg, img_frame, NIR_SENSOR_WIDTH * 3);
  cvNamedWindow(window_name, CV_WINDOW_AUTOSIZE);
  cvShowImage(window_name, cvImg);
  cvWaitKey(30);
  cvReleaseImageHeader(&cvImg);
  free(cvImg);
}

void display_rgb_img(uint8_t *img_frame, const char *window_name)
{
  IplImage *cvImg = cvCreateImageHeader(cvSize(RGB_SENSOR_WIDTH, RGB_SENSOR_HEIGHT), IPL_DEPTH_8U, 3);
  cvSetData(cvImg, img_frame, RGB_SENSOR_WIDTH * 3);
  cvNamedWindow(window_name, CV_WINDOW_AUTOSIZE);
  cvShowImage(window_name, cvImg);
  cvWaitKey(30);
  cvReleaseImageHeader(&cvImg);
  free(cvImg);
}

int main()
{
  nir_tmp = (uint8_t *)malloc(NIR_SENSOR_FRAME_SIZE * sizeof(uint8_t));
  depth_tmp = (uint8_t *)malloc(NIR_SENSOR_FRAME_SIZE * sizeof(uint8_t));
  dot_tmp = (uint8_t *)malloc(NIR_SENSOR_FRAME_SIZE * sizeof(uint8_t));
  rgb_tmp = (uint8_t *)malloc(RGB_SENSOR_FRAME_SIZE * sizeof(uint8_t));

  pthread_t thread_id1, thread_id2, thread_id3, thread_id4;
  pthread_mutex_init(&nir_lock, NULL);
  pthread_mutex_init(&depth_lock, NULL);
  pthread_mutex_init(&dot_lock, NULL);
  pthread_mutex_init(&rgb_lock, NULL);

  struct sockaddr_in *server_addr1 = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
  struct sockaddr_in *server_addr2 = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
  struct sockaddr_in *server_addr3 = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));
  struct sockaddr_in *server_addr4 = (struct sockaddr_in *)malloc(sizeof(struct sockaddr_in));

  server_addr1->sin_family = AF_INET;
  server_addr1->sin_port = htons(nir_port);
  server_addr1->sin_addr.s_addr = inet_addr(ip);

  if (pthread_create(&thread_id1, NULL, init_server_socket, (void *)server_addr1) < 0)
  {
    perror("could not create thread");
    return 1;
  }

  server_addr2->sin_family = AF_INET;
  server_addr2->sin_port = htons(depth_port);
  server_addr2->sin_addr.s_addr = inet_addr(ip);

  if (pthread_create(&thread_id2, NULL, init_server_socket, (void *)server_addr2) < 0)
  {
    perror("could not create thread");
    return 1;
  }

  server_addr3->sin_family = AF_INET;
  server_addr3->sin_port = htons(dot_port);
  server_addr3->sin_addr.s_addr = inet_addr(ip);

  if (pthread_create(&thread_id3, NULL, init_server_socket, (void *)server_addr3) < 0)
  {
    perror("could not create thread");
    return 1;
  }

  server_addr4->sin_family = AF_INET;
  server_addr4->sin_port = htons(rgb_port);
  server_addr4->sin_addr.s_addr = inet_addr(ip);

  if (pthread_create(&thread_id4, NULL, init_server_socket, (void *)server_addr4) < 0)
  {
    perror("could not create thread");
    return 1;
  }

  while (non_stop)
  {
    if (display_nir == true)
    {
      pthread_mutex_lock(&nir_lock);
      display_nir_img(nir_tmp, "NIR");
      display_nir = false;
      pthread_mutex_unlock(&nir_lock);
    }
    if (display_depth == true)
    {
      pthread_mutex_lock(&depth_lock);
      display_nir_img(depth_tmp, "Depth");
      display_depth = false;
      pthread_mutex_unlock(&depth_lock);
    }
    if (display_dot == true)
    {
      pthread_mutex_lock(&dot_lock);
      display_nir_img(dot_tmp, "Dot");
      display_dot = false;
      pthread_mutex_unlock(&dot_lock);
    }
    if (display_rgb == true)
    {
      pthread_mutex_lock(&rgb_lock);
      display_rgb_img(rgb_tmp, "RGB");
      display_rgb = false;
      pthread_mutex_unlock(&rgb_lock);
    }
  }

  return 0;
}
